@@ -11,6 +11,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/time.h>
+#include <sys/select.h>
 #endif
 
 #include "MPEGsystem.h"
@@ -25,22 +26,22 @@ Uint8 const END_CODE[]          = { 0x00, 0x00, 0x01, 0xb9 };
 Uint8 const END_MASK[]          = { 0xff, 0xff, 0xff, 0xff };
 Uint8 const END2_CODE[]         = { 0x00, 0x00, 0x01, 0xb7 };
 Uint8 const END2_MASK[]         = { 0xff, 0xff, 0xff, 0xff };
-Uint8 const VIDEO_CODE[]        = { 0x00, 0x00, 0x01, 0xb3 };
-Uint8 const VIDEO_MASK[]        = { 0xff, 0xff, 0xff, 0xff };
-Uint8 const AUDIO_CODE[]        = { 0xff, 0xf0, 0x00, 0x00 };
-Uint8 const AUDIO_MASK[]        = { 0xff, 0xf0, 0x00, 0x00 };
 Uint8 const VIDEOSTREAM_CODE[]  = { 0x00, 0x00, 0x01, 0xe0 };
 Uint8 const VIDEOSTREAM_MASK[]  = { 0xff, 0xff, 0xff, 0xe0 };
 Uint8 const AUDIOSTREAM_CODE[]  = { 0x00, 0x00, 0x01, 0xc0 };
 Uint8 const AUDIOSTREAM_MASK[]  = { 0xff, 0xff, 0xff, 0xc0 };
-Uint8 const PAD_CODE[]          = { 0x00, 0x00, 0x01, 0xbe };
-Uint8 const PAD_MASK[]          = { 0xff, 0xff, 0xff, 0xff };
+Uint8 const PADSTREAM_CODE[]    = { 0x00, 0x00, 0x01, 0xbe };
+Uint8 const PADSTREAM_MASK[]    = { 0xff, 0xff, 0xff, 0xff };
 Uint8 const SYSTEMSTREAM_CODE[] = { 0x00, 0x00, 0x01, 0xbb };
 Uint8 const SYSTEMSTREAM_MASK[] = { 0xff, 0xff, 0xff, 0xff };
+Uint8 const USERSTREAM_CODE[]   = { 0x00, 0x00, 0x01, 0xb2 };
+Uint8 const USERSTREAM_MASK[]   = { 0xff, 0xff, 0xff, 0xff };
+Uint8 const VIDEO_CODE[]        = { 0x00, 0x00, 0x01, 0xb3 };
+Uint8 const VIDEO_MASK[]        = { 0xff, 0xff, 0xff, 0xff };
+Uint8 const AUDIO_CODE[]        = { 0xff, 0xf0, 0x00, 0x00 };
+Uint8 const AUDIO_MASK[]        = { 0xff, 0xf0, 0x00, 0x00 };
 Uint8 const GOP_CODE[]          = { 0x00, 0x00, 0x01, 0xb8 };
 Uint8 const GOP_MASK[]          = { 0xff, 0xff, 0xff, 0xff };
-Uint8 const USER_CODE[]         = { 0x00, 0x00, 0x01, 0xb2 };
-Uint8 const USER_MASK[]         = { 0xff, 0xff, 0xff, 0xff };
 Uint8 const PICTURE_CODE[]      = { 0x00, 0x00, 0x01, 0x00 };
 Uint8 const PICTURE_MASK[]      = { 0xff, 0xff, 0xff, 0xff };
 Uint8 const SLICE_CODE[]        = { 0x00, 0x00, 0x01, 0x01 };
@@ -194,16 +195,16 @@ static inline Uint32 sequence_header(Uint8 * pointer, Uint32 size, double * _fra
   if((header_size+=8) >= size) return(0);
   switch(pointer[7]&0xF)                /*  4 bits of fps */
   {
-    case 1: frametime = 1/23.97; break;
-    case 2: frametime = 1/24.00; break;
-    case 3: frametime = 1/25.00; break;
-    case 4: frametime = 1/29.97; break;
-    case 5: frametime = 1/30.00; break;
-    case 6: frametime = 1/50.00; break;
-    case 7: frametime = 1/59.94; break;
-    case 8: frametime = 1/60.00; break;
-    case 9: frametime = 1/15.00; break;
-    default: frametime = 1/30.00; break;
+    case 1: frametime = 1.0/23.97; break;
+    case 2: frametime = 1.0/24.00; break;
+    case 3: frametime = 1.0/25.00; break;
+    case 4: frametime = 1.0/29.97; break;
+    case 5: frametime = 1.0/30.00; break;
+    case 6: frametime = 1.0/50.00; break;
+    case 7: frametime = 1.0/59.94; break;
+    case 8: frametime = 1.0/60.00; break;
+    case 9: frametime = 1.0/15.00; break;
+    default: frametime = 1.0/30.00; break;
   }
 
   if(_frametime) *_frametime = frametime;
@@ -286,7 +287,7 @@ static inline Uint32 packet_header(Uint8 * pointer, Uint32 size, double * _times
 }
 
 /* Return true if there is a valid stream header at the beginning of pointer */
-static inline Uint32 stream_header(Uint8 * pointer, Uint32 size, Uint32 * _packet_size, Uint8 * _stream_id, double * _stream_timestamp)
+static inline Uint32 stream_header(Uint8 * pointer, Uint32 size, Uint32 * _packet_size, Uint8 * _stream_id, double * _stream_timestamp, double timestamp)
 {
   Uint32 header_size, packet_size;
   Uint8 stream_id;
@@ -296,8 +297,10 @@ static inline Uint32 stream_header(Uint8 * pointer, Uint32 size, Uint32 * _packe
   if((header_size += 4) >= size) return(0); 
 
   if(!Match4(pointer, SYSTEMSTREAM_CODE, SYSTEMSTREAM_MASK) &&
-     !Match4(pointer, AUDIOSTREAM_CODE, AUDIOSTREAM_MASK) &&
-     !Match4(pointer, VIDEOSTREAM_CODE, VIDEOSTREAM_MASK))
+     !Match4(pointer, AUDIOSTREAM_CODE, AUDIOSTREAM_MASK)   &&
+     !Match4(pointer, VIDEOSTREAM_CODE, VIDEOSTREAM_MASK)   &&
+     !Match4(pointer, PADSTREAM_CODE, PADSTREAM_MASK)       &&
+     !Match4(pointer, USERSTREAM_CODE, USERSTREAM_MASK))
     return(0); /* Unknown encapsulated stream */
 
   /* Parse the stream packet */
@@ -336,7 +339,7 @@ static inline Uint32 stream_header(Uint8 * pointer, Uint32 size, Uint32 * _packe
   else if ( pointer[0] != 0x0f && pointer[0] != 0x80)
       return(0);      /* not a valid header */
   else
-      stream_timestamp = -1;
+      stream_timestamp = timestamp;
     
   if((++header_size) >= size) return(0); 
   --packet_size;
@@ -358,7 +361,7 @@ static inline bool system_aligned(Uint8 *pointer, Uint32 size)
     while((s = packet_header(pointer+i, size-i, 0)) != 0)
       if((i+=s) >= size) return(true);
 
-    if((s = stream_header(pointer+i, size-i, 0, 0, 0)) != 0)
+    if((s = stream_header(pointer+i, size-i, 0, 0, 0, 0)) != 0)
       return(true);
     else
       return(false);
@@ -387,9 +390,9 @@ Uint32 skip_zeros(Uint8 * pointer, Uint32 size)
   return(header_size);
 }
 
-MPEGsystem::MPEGsystem(int Mpeg_FD)
+MPEGsystem::MPEGsystem(SDL_RWops *mpeg_source)
 {
-  mpeg_fd = Mpeg_FD;
+  source = mpeg_source;
 
   /* Create a new buffer for reading */
   read_buffer = new Uint8[MPEG_BUFFER_SIZE];
@@ -398,13 +401,6 @@ MPEGsystem::MPEGsystem(int Mpeg_FD)
   system_mutex = SDL_CreateMutex();
   request_wait = SDL_CreateSemaphore(0);
 
-  /*
-     We are not being passed data, so we don't need to simulate
-     file access.  Might something like MikMod's abstraction of
-     read/seek/tell be appropriate?
-   */
-  data_reader.fromData = false;
-
   /* Invalidate the read buffer */
   pointer = read_buffer;
   read_size = 0;
@@ -450,92 +446,26 @@ MPEGsystem::MPEGsystem(int Mpeg_FD)
 #endif
 
   /* Look for streams */
+  int tries = 0;
   do
   {
     RequestBuffer();
     Wait();
+    if ( tries++ < 10 ) {
+      if ( exist_stream(VIDEO_STREAMID, 0xF0) &&
+	   exist_stream(AUDIO_STREAMID, 0xF0) ) {
+        break;
+      }
+    } else {
+      if ( exist_stream(VIDEO_STREAMID, 0xF0) ||
+	   exist_stream(AUDIO_STREAMID, 0xF0) ) {
+        break;
+      }
+    }
   }
-  while(!exist_stream(VIDEO_STREAMID, 0xF0) &&
-	!exist_stream(AUDIO_STREAMID, 0xF0) &&
-	!Eof());
+  while(!Eof());
 }
 
-MPEGsystem::MPEGsystem(void *data, int size)
-{
-  /* Create a new buffer for reading */
-  read_buffer = new Uint8[MPEG_BUFFER_SIZE];
-
-  /* Create a mutex to avoid concurrent access to the stream */
-  system_mutex = SDL_CreateMutex();
-  request_wait = SDL_CreateSemaphore(0);
-
-  /* 
-     Our argument list indicates that we are being passed data, 
-     so we copy it and setup the data_reader data structure.
-    */
-  data_reader.fromData = true;
-  data_reader.data = malloc(size);
-  data_reader.size = size;
-  data_reader.offset = 0;
-
-  memcpy(data_reader.data, data, size);
-
-  /* Invalidate the read buffer */
-  pointer = read_buffer;
-  read_size = 0;
-  read_total = 0;
-  packet_total = 0;
-  endofstream = errorstream = false;
-  looping = false;
-  frametime = 0.0;
-  stream_timestamp = 0.0;
-
-  /* Create an empty stream list */
-  stream_list = 
-    (MPEGstream **) malloc(sizeof(MPEGstream *));
-  stream_list[0] = 0;
-
-  /* Create the system stream and add it to the list */
-  if(!get_stream(SYSTEM_STREAMID))
-    add_stream(new MPEGstream(this, SYSTEM_STREAMID));
-
-  timestamp = 0.0;
-  timedrift = 0.0;
-  skip_timestamp = -1;
-  system_thread_running = false;
-  system_thread = 0;
-
-  /* Search the MPEG for the first header */
-  if(!seek_first_header())
-  {
-    errorstream = true;
-    SetError("Could not find the beginning of MPEG data\n");
-    return;
-  }
-
-#ifdef USE_SYSTEM_THREAD
-  /* Start the system thread */
-  system_thread = SDL_CreateThread(SystemThread, this);
-
-  /* Wait for the thread to start */
-  while(!system_thread_running && !Eof())
-    SDL_Delay(1);
-#else
-  system_thread_running = true;
-#endif
-
-  /* Look for streams */
-  do
-  {
-    RequestBuffer();
-    Wait();
-  }
-  while(!exist_stream(VIDEO_STREAMID, 0xF0) &&
-	!exist_stream(AUDIO_STREAMID, 0xF0) &&
-	!Eof());
-}
-
-
 MPEGsystem::~MPEGsystem()
 {
   MPEGstream ** list;
@@ -586,77 +516,47 @@ void MPEGsystem::Read()
     /* Replace unread data at the beginning of the stream */
     memmove(read_buffer, pointer, remaining);
 
-    if(data_reader.fromData == true)
-    {
-      /*
-         We were initialized from a chunk of data, so we simulate
-	 the read call
-       */
-
-       int size = READ_ALIGN(MPEG_BUFFER_SIZE - remaining);
-       char *frombuf = (char *) data_reader.data;
-
-       if(data_reader.offset > data_reader.size)
-       {
-         read_size = 0;
-       }
-       else 
-       {
-         if(data_reader.offset + size > data_reader.size)
-         {
-           size = data_reader.size - data_reader.offset;
-         }
-
-         frombuf += data_reader.offset;
-
-         memcpy(read_buffer + remaining, frombuf, size);
-
-         data_reader.offset += size;
-         read_size = size;
-       }
-    } else {
 #ifdef NO_GRIFF_MODS
-        read_size = read(mpeg_fd,
-                        read_buffer + remaining, 
-                        READ_ALIGN(MPEG_BUFFER_SIZE - remaining));
-	if(read_size < 0)
-	{
-	  perror("Read");
-	  errorstream = true;
-	  SDL_mutexV(system_mutex);
-	  return;
-	}
-#else
-	/* Read new data */
-	int bytes_read    = 0;
-	int buffer_offset = remaining;
-	int bytes_to_read = READ_ALIGN(MPEG_BUFFER_SIZE - remaining);
-	int read_at_once = 0;
-
-	read_size = 0;
-	do
-	{
-	  read_at_once = 
-	    read(mpeg_fd, read_buffer + buffer_offset, bytes_to_read );
-
-	  if(read_at_once < 0)
-	  {
-	    perror("Read");
-	    errorstream = true;
-	    SDL_mutexV(system_mutex);
-	    return;
-	  }
-	  else
-	  {
-	    bytes_read    += read_at_once;
-	    buffer_offset += read_at_once;
-	    read_size     += read_at_once;
-	    bytes_to_read -= read_at_once;
-	  }
-	}
-	while( read_at_once>0 && bytes_to_read>0 );
-#endif
+    read_size = SDL_RWread(source,
+                    read_buffer + remaining, 
+                    READ_ALIGN(MPEG_BUFFER_SIZE - remaining));
+    if(read_size < 0)
+    {
+      perror("Read");
+      errorstream = true;
+      SDL_mutexV(system_mutex);
+      return;
     }
+#else
+    /* Read new data */
+    int bytes_read    = 0;
+    int buffer_offset = remaining;
+    int bytes_to_read = READ_ALIGN(MPEG_BUFFER_SIZE - remaining);
+    int read_at_once = 0;
+
+    read_size = 0;
+    do
+    {
+      read_at_once = 
+        SDL_RWread(source, read_buffer + buffer_offset, 1, bytes_to_read );
+
+      if(read_at_once < 0)
+      {
+        perror("Read");
+        errorstream = true;
+        SDL_mutexV(system_mutex);
+        return;
+      }
+      else
+      {
+        bytes_read    += read_at_once;
+        buffer_offset += read_at_once;
+        read_size     += read_at_once;
+        bytes_to_read -= read_at_once;
+      }
+    }
+    while( read_at_once>0 && bytes_to_read>0 );
+#endif
 
     read_total += read_size;
 
@@ -717,7 +617,7 @@ Uint8 MPEGsystem::FillBuffer()
 #endif
   }
 
-  if((header_size = stream_header(pointer, read_buffer + read_size - pointer, &packet_size, &stream_id, &stream_timestamp)) != 0)
+  if((header_size = stream_header(pointer, read_buffer + read_size - pointer, &packet_size, &stream_id, &stream_timestamp, timestamp)) != 0)
   {
       pointer += header_size;
       stream_list[0]->pos += header_size;
@@ -731,6 +631,9 @@ header_size, packet_size, stream_id, stream_timestamp);
      Match4(pointer, END2_CODE, END2_MASK))
   {
     /* End codes belong to video stream */
+#ifdef DEBUG_SYSTEM
+    fprintf(stderr, "[%d] MPEG end code\n", read_total - read_size + (pointer - read_buffer));
+#endif
     stream_id = exist_stream(VIDEO_STREAMID, 0xF0);
     packet_size = 4;
   }
@@ -907,10 +810,10 @@ header_size, packet_size, stream_id, stream_timestamp);
       while (pointer[0] & 0x80 )
       {
 	/* If the stream doesn't already exist */
-	if(!get_stream(pointer[1]))
+	if(!get_stream(pointer[0]))
 	{
 	  /* Create a new stream and add it to the list */
-	  add_stream(new MPEGstream(this, pointer[1]));
+	  add_stream(new MPEGstream(this, pointer[0]));
 	}
 	pointer += 3;
 	stream_list[0]->pos += 3;
@@ -1001,21 +904,12 @@ Uint32 MPEGsystem::TotalSize()
   off_t size;
   off_t pos;
 
-  if(data_reader.fromData == true)
-  {
-    /*
-       we are using simulated data access routines, so we know in advance
-       what the size is.
-     */
-    return data_reader.size;
-  }
-
   /* Lock to avoid concurrent access to the stream */
   SDL_mutexP(system_mutex);
 
   /* I made it this way (instead of fstat) to avoid #ifdef WIN32 everywhere */
   /* in case 'in some weird perversion of nature' someone wants to port this to Win32 :-) */
-  if((pos = lseek(mpeg_fd, 0, SEEK_CUR)) < 0)
+  if((pos = SDL_RWtell(source)) < 0)
   {
     if(errno != ESPIPE)
     {
@@ -1026,7 +920,7 @@ Uint32 MPEGsystem::TotalSize()
     return(0);
   }
 
-  if((size = lseek(mpeg_fd, 0, SEEK_END)) < 0)
+  if((size = SDL_RWseek(source, 0, SEEK_END)) < 0)
   {
     if(errno != ESPIPE)
     {
@@ -1037,7 +931,7 @@ Uint32 MPEGsystem::TotalSize()
     return(0);
   }
   
-  if((pos = lseek(mpeg_fd, pos, SEEK_SET)) < 0)
+  if((pos = SDL_RWseek(source, pos, SEEK_SET)) < 0)
   {
     if(errno != ESPIPE)
     {
@@ -1063,23 +957,20 @@ double MPEGsystem::TotalTime()
   SDL_mutexP(system_mutex);
 
   /* Save current position */
-  if(data_reader.fromData == true) {
-	  /* from data, no lseek needed */
-	  pos = data_reader.offset;
-  } else {
-	  if((pos = lseek(mpeg_fd, 0, SEEK_CUR)) < 0)
-	  {
-	    if(errno != ESPIPE)
-	    {
-	      errorstream = true;
-	      SetError(strerror(errno));
-	    }
-	    SDL_mutexV(system_mutex);
-	    return(false);
-	  }
+  if((pos = SDL_RWtell(source)) < 0)
+  {
+    if(errno != ESPIPE)
+    {
+      errorstream = true;
+      SetError(strerror(errno));
+    }
+    SDL_mutexV(system_mutex);
+    return(false);
   }
+
   file_ptr = 0;
   buffer = new Uint8[MPEG_BUFFER_SIZE];
+  time = 0;
 
   /* If audio, compute total time according to bitrate of the first header and total size */
   /* Note: this doesn't work on variable bitrate streams */
@@ -1087,43 +978,18 @@ double MPEGsystem::TotalTime()
   {
     do
     {
-      if(data_reader.fromData == true) {
-	      data_reader.offset += file_ptr;
-	      size = data_reader.offset;
-      }
-      else {
-	      if((size = lseek(mpeg_fd, file_ptr, SEEK_SET)) < 0)
-	      {
-		if(errno != ESPIPE)
-		{
-		  errorstream = true;
-		  SetError(strerror(errno));  
-		}
-		SDL_mutexV(system_mutex);
-		return(false);
-	      }
+      if((size = SDL_RWseek(source, file_ptr, SEEK_SET)) < 0)
+      {
+        if(errno != ESPIPE)
+        {
+          errorstream = true;
+          SetError(strerror(errno));  
+        }
+        SDL_mutexV(system_mutex);
+        return(false);
       }
     
-      if(data_reader.fromData == true) {
-	      char *dataptr = (char *) data_reader.data;
-	      int size = MPEG_BUFFER_SIZE;
-
-	      if(data_reader.offset + size > data_reader.size) {
-                  size = data_reader.size - data_reader.offset;
-	      }
-
-              if(size >= 0) {
-                  dataptr += size;
-      
-                  memcpy(buffer, dataptr, size);
-      
-                  data_reader.offset += size;
-             } else {
-		     break;
-	     }
-      } else {
-        if(read(mpeg_fd, buffer, MPEG_BUFFER_SIZE) < 0) break;
-      }
+      if(SDL_RWread(source, buffer, 1, MPEG_BUFFER_SIZE) < 0) break;
 
       /* Search for a valid audio header */
       for(p = buffer; p < buffer + MPEG_BUFFER_SIZE; p++)
@@ -1142,43 +1008,31 @@ double MPEGsystem::TotalTime()
     totalsize = TotalSize();
     if(framesize)
       time = frametime * totalsize / framesize;
-    else
-      time = 0;
   }
   else
   {
+    bool last_chance = false;
     do
     {
     /* Otherwise search the stream backwards for a valid header */
       file_ptr -= MPEG_BUFFER_SIZE;
-      
-      if(data_reader.fromData == true) {
-	      data_reader.offset = data_reader.size - file_ptr;
-	      size = data_reader.offset;
-      } else {
-	      if((size = lseek(mpeg_fd, file_ptr, SEEK_END)) < 0)
-	      {
-		if(errno != ESPIPE)
-		{
-		  errorstream = true;
-		  SetError(strerror(errno));  
-		}
-		SDL_mutexV(system_mutex);
-		return(false);
-	      }
+      if ( file_ptr < -(Sint32)TotalSize() ) {
+          last_chance = true;
+          file_ptr = -(Sint32)TotalSize();
+      }
+     
+      if((size = SDL_RWseek(source, file_ptr, SEEK_END)) < 0)
+      {
+        if(errno != ESPIPE)
+        {
+          errorstream = true;
+          SetError(strerror(errno));  
+        }
+        SDL_mutexV(system_mutex);
+        return(false);
       }
       
-      if(data_reader.fromData == true) {
-        char *dataptr = (char *) data_reader.data;
-
-        dataptr += data_reader.offset;
-
-        memcpy(buffer, dataptr, MPEG_BUFFER_SIZE);
-
-        data_reader.offset += MPEG_BUFFER_SIZE;
-      } else {
-        if(read(mpeg_fd, buffer, MPEG_BUFFER_SIZE) < 0) break;
-      }
+      if(SDL_RWread(source, buffer, 1, MPEG_BUFFER_SIZE) < 0) break;
       
       if(stream_list[0]->streamid == SYSTEM_STREAMID)
 	for(p = buffer + MPEG_BUFFER_SIZE - 1; p >= buffer;)
@@ -1187,6 +1041,7 @@ double MPEGsystem::TotalTime()
 	  if(*p-- != 1) continue;
 	  if(*p-- != 0) continue;
 	  if(*p-- != 0) continue;
+          ++p;
 	  break;
 	}
       if(stream_list[0]->streamid == VIDEO_STREAMID)
@@ -1196,28 +1051,77 @@ double MPEGsystem::TotalTime()
 	  if(*p-- != 1) continue;
 	  if(*p-- != 0) continue;
 	  if(*p-- != 0) continue;
+          ++p;
 	  break;
 	}
     }
-    while(p < buffer);
+    while( !last_chance && (p < buffer) );
 
-    p++;
-
-    /* Extract time info from the last header */
-    if(stream_list[0]->streamid == SYSTEM_STREAMID)
-      packet_header(p, buffer + MPEG_BUFFER_SIZE - p, &time);
+    if ( p >= buffer ) {
+      /* Extract time info from the last header */
+      if(stream_list[0]->streamid == SYSTEM_STREAMID)
+        packet_header(p, buffer + MPEG_BUFFER_SIZE - p, &time);
     
-    if(stream_list[0]->streamid == VIDEO_STREAMID)
-      gop_header(p, buffer + MPEG_BUFFER_SIZE - p, &time);
+      if(stream_list[0]->streamid == VIDEO_STREAMID)
+        gop_header(p, buffer + MPEG_BUFFER_SIZE - p, &time);
+    }
   }
 
-  delete buffer;
+  delete[] buffer;
 
-  if(data_reader.fromData == true) {
-      data_reader.offset = pos;
-  } else {
-      /* Get back to saved position */
-      if((pos = lseek(mpeg_fd, pos, SEEK_SET)) < 0)
+  /* Get back to saved position */
+  if((pos = SDL_RWseek(source, pos, SEEK_SET)) < 0)
+  {
+    if(errno != ESPIPE)
+    {
+      errorstream = true;
+      SetError(strerror(errno));  
+    }
+    time = 0;
+  }
+
+  SDL_mutexV(system_mutex);
+
+  return(time);
+}
+
+double MPEGsystem::TimeElapsedAudio(int atByte)
+{
+  off_t size, pos;
+  off_t file_ptr;
+  Uint8 * buffer, * p;
+  double time;
+  
+  if (atByte < 0)
+  {
+      return -1;
+  }
+  
+  /* Lock to avoid concurrent access to the stream */
+  SDL_mutexP(system_mutex);
+
+  /* Save current position */
+  if((pos = SDL_RWtell(source)) < 0)
+  {
+    if(errno != ESPIPE)
+    {
+      errorstream = true;
+      SetError(strerror(errno));
+    }
+    SDL_mutexV(system_mutex);
+    return(false);
+  }
+
+  file_ptr = 0;
+  buffer = new Uint8[MPEG_BUFFER_SIZE];
+
+  /* If audio, compute total time according to bitrate of the first header and total size */
+  /* Note: this doesn't work on variable bitrate streams */
+  if(stream_list[0]->streamid == AUDIO_STREAMID)
+  {
+    do
+    {
+      if((size = SDL_RWseek(source, file_ptr, SEEK_SET)) < 0)
       {
         if(errno != ESPIPE)
         {
@@ -1225,12 +1129,53 @@ double MPEGsystem::TotalTime()
           SetError(strerror(errno));  
         }
         SDL_mutexV(system_mutex);
-        return(0);
+        return(false);
       }
+    
+      if(SDL_RWread(source, buffer, 1, MPEG_BUFFER_SIZE) < 0) break;
+
+      /* Search for a valid audio header */
+      for(p = buffer; p < buffer + MPEG_BUFFER_SIZE; p++)
+	if(audio_aligned(p, buffer + MPEG_BUFFER_SIZE - p)) break;
+    
+      file_ptr += MPEG_BUFFER_SIZE;
+    }
+    while(p >= MPEG_BUFFER_SIZE + buffer);
+
+    /* Extract time info from the first header */
+    Uint32 framesize;
+    double frametime;
+    Uint32 totalsize;
+
+    audio_header(p, &framesize, &frametime);
+    totalsize = TotalSize();
+    if(framesize)
+      //is there a better way to do this?
+      time = (frametime * (atByte ? atByte:totalsize)) / framesize;
+    else
+      time = 0;
+  }
+  else
+  //This is not a purely audio stream. This doesn't make sense!
+  {
+      time = -1;
+  }
+
+  delete buffer;
+
+  /* Get back to saved position */
+  if((pos = SDL_RWseek(source, pos, SEEK_SET)) < 0)
+  {
+    if(errno != ESPIPE)
+    {
+      errorstream = true;
+      SetError(strerror(errno));  
+    }
+    SDL_mutexV(system_mutex);
+    return(0);
   }
 
   SDL_mutexV(system_mutex);
-
   return(time);
 }
 
@@ -1247,19 +1192,15 @@ bool MPEGsystem::Seek(int length)
   /* Lock to avoid concurrent access to the stream */
   SDL_mutexP(system_mutex);
   
-  if(data_reader.fromData == true) {
-    data_reader.offset = length;
-  } else {
-    /* Get into the stream */
-    if(lseek(mpeg_fd, length, SEEK_SET) < 0)
+  /* Get into the stream */
+  if(SDL_RWseek(source, length, SEEK_SET) < 0)
+  {
+    if(errno != ESPIPE)
     {
-      if(errno != ESPIPE)
-      {
-        errorstream = true;
-        SetError(strerror(errno));
-      }
-      return(false);
+      errorstream = true;
+      SetError(strerror(errno));
     }
+    return(false);
   }
 
   /* Reinitialize the read buffer */
@@ -1359,20 +1300,15 @@ bool MPEGsystem::SystemLoop(MPEGsystem *system)
     /* Set the eof mark on all streams */
     system->end_all_streams();
 
-    if(system->data_reader.fromData == true)
+    /* Get back to the beginning of the stream if possible */
+    if(SDL_RWseek(system->source, 0, SEEK_SET) < 0)
     {
-      system->data_reader.offset = 0;
-    } else {
-      /* Get back to the beginning of the stream if possible */
-      if(lseek(system->mpeg_fd, 0, SEEK_SET) < 0)
+      if(errno != ESPIPE)
       {
-	if(errno != ESPIPE)
-	{
-	  system->errorstream = true;
-	  system->SetError(strerror(errno));
-	}
-        return(false);
+        system->errorstream = true;
+        system->SetError(strerror(errno));
       }
+      return(false);
     }
 
     /* Reinitialize the read buffer */
